structure Drive =

struct

structure A = AbstractSyntax;

val top_env:Semant.environment ref = ref(Semant.empty_env);

val top_type_env:Types.type_env ref = ref([]);

val top_ab:AssumptionBase.assum_base ref = ref(AssumptionBase.empty_ab);

val (no_types,min_out) = (ref false, ref false); (* No types, minimal output *)

fun getResultAndComplexities(phr,env,ab) = 
     let val (v,c) = Semant.vcPhrase(phr,env,ab)
         val (_,ec) = Semant.vc_E_Phrase(phr,env,ab)
         val (_,dc) = Semant.vc_D_Phrase(phr,env,ab)
     in
        (v,c,ec,dc)
     end;

fun printComp(c,ec,dc) = 
        (print("\nTotal complexity: "^Int.toString(c)^"\n");
         print("Computational complexity: "^Int.toString(ec)^"\n");
         print("Deductive complexity: "^Int.toString(dc)^"\n"));

fun printCert(str,c,ec,dc) = 
         (print("\nAnd here is the certificate: \n\n"^str^"\n");
          print("\nThe total complexity of the certificate was: "^Int.toString(c)^"\n");
          print("Its computational complexity was: "^Int.toString(ec)^"\n");
          print("And its deductive complexity: "^Int.toString(dc)^"\n"));
          
fun pt(SOME(t)) = Types.printType(t)
  | pt(_) = "";

fun processPhrase(ph as A.expression(e)) = 
      let val t = if !no_types then NONE else SOME(Types.inferType(ph,!top_type_env))
          val v = Semant.evExp(e,top_env,!top_ab)
          val (_,c,ec,dc) = getResultAndComplexities(A.expression(e),top_env,!top_ab)
      in 
          (if !no_types then print("\n"^Semant.printVal(v)^"\n")
           else print("\n"^Semant.printVal(v)^": "^pt(t)^"\n");
           if !min_out then () else printComp(c,ec,dc))
      end
  | processPhrase(ph as A.deduction(d)) = 
      let val _ = if !no_types then NONE else SOME(Types.inferType(ph,!top_type_env))
          val th as Semant.propVal(P) = Semant.evDed(d,top_env,!top_ab)
      in 
         (print("\nTheorem: "^Semant.printVal(th)^"\n");
          if !min_out then ()
          else let val (_,c,ec,dc) = getResultAndComplexities(ph,top_env,!top_ab)
                   val (v as Semant.propVal(P),SOME(D)) = Semant.evDedAndExpand(d,top_env,!top_ab)
                   val dstr = A.unparseAlphaDed(A.lift(D))
                   val (_,expanded_cost,expanded_e_cost,expanded_d_cost) = 
                         getResultAndComplexities(A.deduction(D),top_env,!top_ab)
               in
                 (printComp(c,ec,dc);printCert(dstr,expanded_cost,expanded_e_cost,expanded_d_cost))
               end;
          top_ab := AssumptionBase.insert(P,!top_ab))
      end;

fun processDefinition(str,phr) = let val gt_opt = if !no_types then NONE else
                                                   let val t = Types.inferType(phr,!top_type_env)
                                                   in
                                                     SOME(Types.generalize(t,Basic.removeDuplicatesEq(
                                                          Types.freeTVars(t),op=)))
                                                   end
                                     val v = Semant.evPhrase(phr,top_env,!top_ab)
                                 in
                                    (top_env := Semant.extend(!top_env,[(str,Semant.bind(v))]);
                                     (case gt_opt of 
                                         SOME(gt) => 
                                             (top_type_env := Types.extendTypeEnv(!top_type_env,str,gt);
                                              print("\n"^str^": "^Types.printType(gt)^"\n"))
                                       | _ => ()))
                                 end;

fun processAssertion(e) = (case Semant.evExp(e,top_env,!top_ab) of 
                             Semant.propVal(P) => top_ab := AssumptionBase.insert(P,!top_ab));

fun processFile(file_name,f) = 
                    let val (user_input_lst_from_file,msg) = 
                                 (Parse.parse(file_name),"")
                                        handle IO.Io(_) => ([],"Unable to open "^file_name^".")
                     in 
                        (if msg = "" then (List.app (fn i => f(i,file_name))
                                                    user_input_lst_from_file)
                         else print("\n"^msg^"\n"))
                     end

fun processInput(user_input,file) = 
           (case user_input of
               A.phraseInput(some_phrase) => 
                   (processPhrase(some_phrase) 
                       handle Types.UnifyError => print("\nType error.\n")
                            | Types.UnBoundIdentifier => print("\nUnbound identifier.\n"))
             | A.definition(str,phr) => 
                   (processDefinition(str,phr)
                       handle Types.UnifyError => print("\nType error.\n")
                            | Types.UnBoundIdentifier => print("\nUnbound identifier.\n"))
             | A.assertion(e) => (processAssertion(e);print("\nOK.\n"))
             | A.minimalPrintOuts => (min_out := true; print("\nOK.\n"))
             | A.noTypes => (no_types := true; print("\nOK.\n"))
             | A.load(str) => (print("\nFile to be opened: "^str^"\n");processFile(str,processInput)));

fun killNDL() = OS.Process.terminate(OS.Process.success);

fun test() =
    let val double_quote_code = 34
        val back_slash_code = 92
        val first_time = ref(true)
        fun stringToChar(s) = (case Char.fromString(s) of SOME(c) => c | _ => #"a")
        fun isDoubleQuote(str) = (ord(stringToChar(str)) = double_quote_code)
        fun isBackSlash(str) = ord(stringToChar(str)) = back_slash_code
        fun boolToString(b) = if b = true then "TRUE" else "FALSE"
        fun balancedParentheses(str) = 
          let val str_lst = map Char.toString (String.explode(str))
              fun printable(str) = (case Char.fromString(str) of
                                      SOME(c) => if Int.>(Char.ord(c),32) then true else false
                                    | _ => false)
              fun whiteSpace(str) = not(printable(str))
              fun findFirstWhiteSpaceOrRightDelim([]) = []
                | findFirstWhiteSpaceOrRightDelim(str as x::more) = 
                       if whiteSpace(x) orelse x = ")" orelse x = "]"
                       then str else
                       findFirstWhiteSpaceOrRightDelim(more)
              fun findFirstPrintable([]) = []
                | findFirstPrintable(str as x::more) = 
                       if not(whiteSpace(x))
                       then str else
                       findFirstPrintable(more)
              fun findFirstNewLine([]) = []
                | findFirstNewLine(x::more) = 
                        if x = "\\n" then more
                      else findFirstNewLine(more)
              fun findFirstDoubleQuote([]) = []
                | findFirstDoubleQuote(str as x::y::more) = 
                    if isBackSlash(x) then
                         findFirstDoubleQuote(more)
                    else
                       if isDoubleQuote(x) then y::more
                       else
                          findFirstDoubleQuote(y::more)
                | findFirstDoubleQuote(str as x::more) = 
                     if isDoubleQuote(x) then more
                     else 
                         findFirstDoubleQuote(more)
              fun canStartId(str) = printable(str) andalso
                   not(Basic.isMember(str,["?","'","!","`","$","(",")","[","]",";","\""]))
                   andalso not(isDoubleQuote(str));
              fun bp([],lp,rp,lb,rb) = ((lp = rp) andalso (lb = rb)) orelse
                                        (rp > lp) orelse (rb > lb)
                | bp(x::more,lp,rp,lb,rb) = 
                      if whiteSpace(x) then
                         bp(more,lp,rp,lb,rb)
                      else
                         if canStartId(x) then
                            bp(findFirstWhiteSpaceOrRightDelim(more),lp,rp,lb,rb)
                         else 
                            if isDoubleQuote(x) then
                               bp(findFirstDoubleQuote(more),lp,rp,lb,rb)
                         else
                            (case x of 
                                "(" => bp(more,lp+1,rp,lb,rb)
                              | "[" => bp(more,lp,rp,lb+1,rb)
                              | ")" => bp(more,lp,rp+1,lb,rb)
                              | "]" => bp(more,lp,rp,lb,rb+1)
                              | ";" => bp(findFirstNewLine(more),lp,rp,lb,rb)
                              | _ => bp(more,lp,rp,lb,rb))
             val open_string = ref(false)
          in
             let val p = findFirstPrintable(str_lst)
             in
                if null(p) then false
                else
                   if isDoubleQuote(hd(p)) then
                      not(null(findFirstDoubleQuote(tl(p))))
                   else
                      if Basic.isMember(hd(p),["(","[",";"]) then 
                         bp(str_lst,0,0,0,0)
                      else true
             end
          end
        val in_str = Basic.readSexp("\n>",balancedParentheses) 
        val in_str' = if !first_time then 
                         let val _ = first_time := false 
                             val str_lst = explode(in_str)
                         in
                           (case explode(in_str) of
                               #"\n"::more => implode(more)
                             | _ => in_str)
                         end
                      else
                         in_str 
    in
       if in_str' = "quit\n" then ()
       else if in_str' = "" then killNDL() 
       else
          let val user_input =
                      SOME(hd(Parse.parse_from_stream(TextIO.openString(in_str'))))
                      handle ErrorMsg.Error => NONE
          in
            (case user_input of 
               SOME(user_in) => (processInput(user_in,"top level");test())
              | _ => test())
          end
    end;


end; (** Of structure Drive **)

