signature ERRORMSG =
sig
    val anyErrors : bool ref
    val fileName : string ref
    val lineNum : int ref
    val linePos : int list ref
    val sourceStream : TextIO.instream ref
    val error: string * (int * int) * (int * int) -> unit
    val error1 : int -> string -> unit
    val print_error: string * (int * int) * (int * int) -> unit    
    exception Error
    exception ParseError of (int * int) * string 
    val impossible : string -> 'a   (* raises Error *)
    val reset : unit -> unit
end

structure ErrorMsg : ERRORMSG =

struct

  exception ParseError of (int * int) * string;

  val anyErrors = ref false
  val fileName = ref ""
  val lineNum = ref 1
  val linePos = ref [1]
  val sourceStream = ref TextIO.stdIn

  fun reset() = (anyErrors:=false;
                 fileName:="";
                 lineNum:=1;
                 linePos:=[1];
                 sourceStream:=TextIO.stdIn)

  exception Error
 
  fun error(str,left_pos as (l1,p1),right_pos as (l2,p2)) = raise ParseError((l1,p1),str);

  fun error1 pos (msg:string) =
      let fun look(a::rest,n) =
                if a < pos then app print [":",
                                       Int.toString n,
                                       ".",
                                       Int.toString (pos-a)]
                       else look(rest,n-1)
            | look _ = print "0.0"
       in anyErrors := true;
          if not("" = !fileName) then print("\n"^(!fileName)^", ") else print("\n");
          print(Position.posToString(Position.getPos(pos)));
          print ": ";
          print msg;
          print "\n"
      end

  val print_error = error

  fun impossible msg =
      (app print ["Error: ",msg,"\n"];
       TextIO.flushOut TextIO.stdOut;
       raise Error)

end  (* structure ErrorMsg *)

